# ANTS-Hub - Copyright (C) Entr'ouvert

import base64
import binascii
import collections
import datetime
import functools
import json
import logging
import os

import jsonschema
import requests
from django.core.exceptions import ValidationError
from django.db.models import Count
from django.db.transaction import atomic
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from ants_hub.data.models import (
    Collectivite,
    Config,
    Horaire,
    HoraireList,
    Lieu,
    Plage,
    Raccordement,
    RendezVous,
    TypeDeRdv,
)
from ants_hub.timezone import localtime, make_naive, now

from .ants import ANTS_TIMEZONE

logger = logging.getLogger('ants_hub.api.chrono')


with open(os.path.join(os.path.dirname(__file__), '..', 'static', 'schemas', 'rdv-disponibles.json')) as fd:
    RENDEZ_VOUS_DISPONIBLES_SCHEMA = json.load(fd)


class Http401(JsonResponse):
    def __init__(self, message):
        super().__init__({'err': message}, status=401, safe=False)
        self['WWW-Authenticate'] = 'Basic'


def authenticate(func):
    @functools.wraps(func)
    def wrapper(request, *args, **kwargs):
        header = request.headers.get('Authorization', '').encode()
        if not header:
            logger.warning('authentication failed, missing Authorization header')
            return Http401("Missing Authorization header")
        auth = header.split()
        if not auth or auth[0].lower() != b'basic' or len(auth) != 2:
            logger.warning('authentication failed, invalid Authorization header')
            return Http401("Invalid Authorization header")

        try:
            auth_decoded = base64.b64decode(auth[1]).decode()
        except (TypeError, UnicodeDecodeError, binascii.Error):
            logger.warning('authentication failed, invalid Authorization header')
            return Http401("Invalid Authorization header")

        apikey = auth_decoded.split(':', 1)[0]
        raccordement = Raccordement.objects.get_by_apikey(apikey)

        if not raccordement:
            logger.warning('authentication failed, unknown API key "%s"', apikey)
            return Http401("Unknown API key")
        request.raccordement = raccordement
        return func(request, *args, **kwargs)

    return wrapper


@authenticate
def ping(request):
    logger.info('ping received')
    return JsonResponse({'err': 0})


class RendezVousDisponibleView(View):
    def get(self, request):
        collectivites_data = []
        for collectivite in request.raccordement.collectivites.prefetch_related('lieux').order_by('created'):
            lieux_data = []
            collectivite_data = {
                'id': collectivite.source_id,
                'nom': collectivite.nom,
                'url': collectivite.url,
                'created': collectivite.created.isoformat(),
                'last_update': collectivite.created.isoformat(),
                'lieux': lieux_data,
                'nombre_de_lieux': collectivite.lieux.count(),
                'nombre_de_pre_demandes_actives': RendezVous.objects.filter(
                    date__gte=now(), lieu__collectivite=collectivite
                ).count(),
                'nombre_de_jours_avec_rdv_disponibles': Plage.objects.filter(
                    lieu__collectivite=collectivite, date__gte=now().date()
                ).aggregate(Count('date'))['date__count'],
            }
            for key in ['logo_url', 'rdv_url', 'gestion_url', 'annulation_url']:
                if getattr(collectivite, key, None):
                    collectivite_data[key] = getattr(collectivite, key)
            for lieu in collectivite.lieux.order_by('created'):
                lieu_data = {
                    'id': lieu.source_id,
                    'nom': lieu.nom,
                    'numero_rue': lieu.numero_rue,
                    'code_postal': lieu.code_postal,
                    'ville': lieu.ville,
                    'longitude': lieu.longitude,
                    'latitude': lieu.latitude,
                    'nombre_de_pre_demandes_actives': lieu.rdvs.filter(date__gte=now()).count(),
                    'nombre_de_jours_avec_rdv_disponibles': lieu.plages.filter(
                        date__gte=now().date()
                    ).aggregate(Count('date'))['date__count'],
                }
                for key in ['url', 'logo_url', 'rdv_url', 'gestion_url', 'annulation_url']:
                    if getattr(lieu, key, None):
                        lieu_data[key] = getattr(collectivite, key)
                lieux_data.append(lieu_data)
            collectivites_data.append(collectivite_data)

        logger.info(
            'rendez-vous-disponibles(%s) returned %d collectivites',
            request.raccordement,
            len(collectivites_data),
        )
        return JsonResponse(
            {
                'err': 0,
                'collectivites': collectivites_data,
            }
        )

    collectivites_created = 0
    collectivites_updated = 0
    lieux_created = 0
    lieux_updated = 0
    lieux_deleted = 0
    plage_created = 0
    plage_updated = 0
    plage_deleted = 0
    rdv_created = 0
    rdv_updated = 0
    rdv_deleted = 0

    def handle_rdv_payload(self, lieu, rdv):
        # cannot fail, as JSON schema is already validated
        identifiant_predemande = rdv.pop('id').strip()
        date = datetime.datetime.fromisoformat(rdv.pop('date'))
        annule = bool(rdv.pop('annule', False))
        if annule:
            _, count_by_model = lieu.rdvs.filter(
                identifiant_predemande=identifiant_predemande, date=date
            ).delete()
            self.rdv_deleted += count_by_model.get('data.RendezVous', 0)
            return
        try:
            rdv = lieu.rdvs.get(identifiant_predemande=identifiant_predemande, date=date, **rdv)
        except RendezVous.DoesNotExist:
            rdv, created = lieu.rdvs.update_or_create(
                identifiant_predemande=identifiant_predemande, date=date, defaults=rdv
            )
            if created:
                self.rdv_created += 1
            else:
                self.rdv_updated += 1
        return rdv

    def handle_rdvs_payload(self, lieu, rdvs, full=False):
        rdv_pks = set()
        for rdv in rdvs:
            rdv = self.handle_rdv_payload(lieu, rdv)
            if full:
                rdv_pks.add(rdv.pk)
        if full:
            _, count_by_model = lieu.rdvs.exclude(pk__in=rdv_pks).delete()
            self.rdv_deleted += count_by_model.get('data.RendezVous', 0)
        return lieu

    def handle_plages_payload(self, lieu, plages, full=False):
        by_date_and_type_and_personnes = collections.defaultdict(set)
        day_is_full = set()
        for plage in plages:
            try:
                date = datetime.date.fromisoformat(plage['date'])
                types_rdv = set()
                # by JSON schema, types_rdv is never empty and always filled
                # with expected values
                for typ in plage['types_rdv']:
                    types_rdv.add(TypeDeRdv.from_ants_name(typ))
                personnes = plage.get('personnes', 1)
                if not plage.get('heure_debut'):
                    day_is_full.update((date, t, personnes) for t in types_rdv)
                    continue
                heure_debut = datetime.time.fromisoformat(plage['heure_debut'])
                heure_fin = datetime.time.fromisoformat(plage['heure_fin'])
                heure_debut = heure_debut.replace(tzinfo=None)
                heure_fin = heure_fin.replace(tzinfo=None)
                duree = int(plage['duree'])
                if heure_fin < heure_debut:
                    raise ValueError('heure de fin inférieure à heure de début')
                if datetime.datetime.combine(date, heure_fin) - datetime.datetime.combine(
                    date, heure_debut
                ) < datetime.timedelta(minutes=duree):
                    raise ValueError('différence heure de début et de fin inférieure à la durée')
                for t in types_rdv:
                    by_date_and_type_and_personnes[(date, t, personnes)].add((heure_debut, heure_fin, duree))
            except ValueError as e:
                raise ValidationError('plage %s: %s' % (plage, e))

        plage_pks = set()
        for x in set(by_date_and_type_and_personnes) | day_is_full:
            date, type_rdv, personnes = x
            if x in day_is_full:
                _, count_by_model = lieu.plages.filter(
                    date=date, type_de_rdv=type_rdv, personnes=personnes
                ).delete()
                self.plage_deleted += count_by_model.get('data.Plage', 0)
                continue
            values = by_date_and_type_and_personnes.get(x)
            durees = list({duree for heure_debut, heure_fin, duree in values})
            if len(durees) > 1:
                raise ValidationError(
                    'plages %s: la durée varie pour les mêmes rendez-vous du même lieu' % values
                )
            duree = durees[0]
            horaires = HoraireList()
            for start, end, _ in values:
                try:
                    horaires.append(Horaire(start=start, end=end))
                except ValueError:
                    raise ValidationError('plages %s-%s: les périodes se chevauchent' % (date, values))
            try:
                plage = lieu.plages.get(
                    date=date, type_de_rdv=type_rdv, personnes=personnes, horaires=horaires, duree=duree
                )
            except Plage.DoesNotExist:
                plage, created = lieu.plages.update_or_create(
                    date=date,
                    type_de_rdv=type_rdv,
                    personnes=personnes,
                    defaults={'horaires': horaires, 'duree': duree},
                )
                if created:
                    self.plage_created += 1
                else:
                    self.plage_updated += 1
            if full:
                plage_pks.add(plage.pk)
        if full:
            _, count_by_model = lieu.plages.exclude(pk__in=plage_pks).delete()
            self.plage_deleted += count_by_model.get('data.Plage', 0)

    def handle_lieu_payload(self, collectivite, payload):
        source_id = payload.pop('id')
        plages = payload.pop('plages', [])
        rdvs = payload.pop('rdvs', [])
        full = payload.pop('full', False)

        try:
            # fast path, instance exactly match
            lieu = collectivite.lieux.get(source_id=source_id, **payload)
        except Lieu.DoesNotExist:
            # get existing pk and try a full_clean
            pk = collectivite.lieux.filter(source_id=source_id).values_list('pk', flat=True).first()
            Lieu(pk=pk, collectivite=collectivite, source_id=source_id, **payload).full_clean()
            # now update or create
            lieu, created = collectivite.lieux.update_or_create(source_id=source_id, defaults=payload)
            if created:
                self.lieux_created += 1
            else:
                self.lieux_updated += 1

        self.handle_plages_payload(lieu, plages, full=full)
        self.handle_rdvs_payload(lieu, rdvs, full=full)
        return lieu

    def handle_collectivite_payload(self, raccordement, payload):
        source_id = payload.pop('id')
        lieux_payload = payload.pop('lieux', [])
        full = payload.pop('full', False)

        try:
            # fast path, instance exactly match
            collectivite = raccordement.collectivites.get(source_id=source_id, **payload)
        except Collectivite.DoesNotExist:
            # get existing pk and try a full_clean
            pk = raccordement.collectivites.filter(source_id=source_id).values_list('pk', flat=True).first()
            Collectivite(pk=pk, raccordement=raccordement, source_id=source_id, **payload).full_clean()
            # now update or create
            collectivite, created = raccordement.collectivites.update_or_create(
                source_id=source_id, defaults=payload
            )
            if created:
                self.collectivites_created += 1
            else:
                self.collectivites_updated += 1
        lieu_pks = set()
        for lieu_payload in lieux_payload:
            lieu = self.handle_lieu_payload(collectivite, lieu_payload)
            if full:
                lieu_pks.add(lieu.pk)
        if full:
            _, count_by_model = collectivite.lieux.exclude(pk__in=lieu_pks).delete()
            self.lieux_deleted += count_by_model.get('data.Lieu', 0)
            self.plage_deleted += count_by_model.get('data.Plage', 0)
            self.rdv_deleted += count_by_model.get('data.RendezVous', 0)

    def post(self, request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return JsonResponse({'err': 'invalid-json'}, status=400)
        try:
            jsonschema.validate(payload, RENDEZ_VOUS_DISPONIBLES_SCHEMA)
        except jsonschema.ValidationError as e:
            return JsonResponse({'err': 'invalid-json', 'detail': e.message}, status=400)
        try:
            with atomic():
                # prevent concurrent updates to the data of the same raccordement
                request.raccordement.lock()
                for collectivite in payload.get('collectivites', []):
                    self.handle_collectivite_payload(request.raccordement, collectivite)
        except (ValueError, ValidationError) as e:
            logger.warning('rendez-vous-disponibles(%s) received bad request "%s"', request.raccordement, e)
            return JsonResponse(
                {
                    'err': 1,
                    'error': str(e),
                },
                status=400,
            )
        actions = []
        result = {}
        for action in ['created', 'updated', 'deleted']:
            parts = []
            for model in ['collectivites', 'lieux', 'plage', 'rdv']:
                name = f'{model}_{action}'
                if name == 'collectivites_deleted':
                    continue
                result[name] = count = getattr(self, name)
                if count:
                    parts.append(f'{count} {model}')
            if parts:
                actions.append(f'{action} {", ".join(parts)}')
        if actions:
            logger.info('rendez-vous-disponibles(%s) %s', request.raccordement, '; '.join(actions))
        else:
            logger.debug('rendez-vous-disponibles(%s) nothing done.', request.raccordement)
        return JsonResponse(
            {
                'err': 0,
                'data': result,
            }
        )


rendez_vous_disponibles = csrf_exempt(authenticate(RendezVousDisponibleView.as_view()))


class PredemandesView(View):
    def get_autres_demandes(self, identifiant_predemande):
        auth_token = Config.get(Config.REQUEST_TO_ANTS_AUTH_TOKEN)
        url = Config.get(Config.REQUEST_TO_ANTS_BASE_URL, 'https://rendez-vous-api.france-identite.fr/api/')
        if not auth_token:
            logger.error(
                'predemandes(%s) server error REQUEST_TO_ANTS_AUTH_TOKEN is not configured',
                self.request.raccordement,
            )
            return None, JsonResponse({'err': 'auth-token-not-configured'}, status=500)
        try:
            response = requests.get(
                f'{url}searchApplicationIds',
                headers={'x-hub-rdv-auth-token': auth_token},
                params={'applications_ids': identifiant_predemande},
                timeout=10,
            )
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error('predemandes(%s) %s is down, %s', self.request.raccordement, url, e)
            return None, JsonResponse(
                {'err': 1, 'err_class': 'rendez-vous-api-is-unavailable', 'err_desc': str(e)}, status=200
            )
        try:
            data = response.json().get(identifiant_predemande, [])
        except (ValueError, AttributeError) as e:
            logger.error(
                'predemandes(%s) %s is down, result is not a JSON object: %s',
                self.request.raccordement,
                url,
                response.content[:1024],
            )
            return None, JsonResponse(
                {'err': 1, 'err_class': 'rendez-vous-api-is-unavailable', 'err_desc': str(e)}, status=200
            )
        except KeyError:
            data = []
        for rendez_vous in data:
            # fix stupid formatting of dates from ANTS, Paris time with 'Z' suffix :/
            date = rendez_vous.get('datetime', None)
            date = datetime.datetime.fromisoformat(rendez_vous['datetime'].rstrip('Z'))
            date = date.replace(tzinfo=ANTS_TIMEZONE)
            date = localtime(date)
            rendez_vous['datetime'] = make_naive(date).isoformat()
        logger.info(
            'predemandes(%s) returned %s demandes from %s for predemande "%s"',
            self.request.raccordement,
            len(data),
            ','.join(map(str, (rendez_vous['meeting_point'] for rendez_vous in data))),
            identifiant_predemande,
        )
        return data, None

    def get(self, request):
        identifiant_predemande = request.GET.get('identifiant_predemande', '').strip()
        if not identifiant_predemande:
            return JsonResponse({'err': 'missing=identifiant-predemande'}, status=400)
        autres_demandes, error_response = self.get_autres_demandes(identifiant_predemande)
        if error_response is not None:
            return error_response

        return JsonResponse(
            {
                'err': 0,
                'data': autres_demandes,
            }
        )


predemandes = csrf_exempt(authenticate(PredemandesView.as_view()))
